@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1999-2019, University of Amsterdam
+    Copyright (c)  1999-2023, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -39,27 +40,17 @@
             statistics/1,               % -Stats
             thread_statistics/2,        % ?Thread, -Stats
             time/1,                     % :Goal
-            profile/1,                  % :Goal
-            profile/2,                  % :Goal, +Options
-            show_profile/1,             % +Options
-            profile_data/1,             % -Dict
-            profile_procedure_data/2    % :PI, -Data
+            call_time/2,                % :Goal, -Time
+            call_time/3                 % :Goal, -Time, -Result
           ]).
-:- autoload(library(error),[must_be/2]).
-:- autoload(library(lists),[append/3,member/2]).
-:- autoload(library(option),[option/3]).
-:- autoload(library(pairs),[map_list_to_pairs/3,pairs_values/2]).
-:- autoload(library(prolog_code),
-	    [predicate_sort_key/2,predicate_label/2]).
-
+:- autoload(library(pairs),[map_list_to_pairs/3]).
 
 :- set_prolog_flag(generate_debug_info, false).
 
 :- meta_predicate
     time(0),
-    profile(0),
-    profile(0, +),
-    profile_procedure_data(:, -).
+    call_time(0, -, -),
+    call_time(0, -).
 
 /** <module> Get information about resource usage
 
@@ -289,12 +280,13 @@ thread_stats(Thread, Stacks,
 %   @bug Inference statistics are often a few off.
 %   @see statistics/2 for obtaining statistics in your program and
 %        understanding the reported values.
+%   @see call_time/2, call_time/3 to obtain the timing in a dict.
 
 time(Goal) :-
     time_state(State0),
     (   call_cleanup(catch(Goal, E, (report(State0,10), throw(E))),
                      Det = true),
-        time_true(State0),
+        time_true_report(State0),
         (   Det == true
         ->  !
         ;   true
@@ -303,281 +295,96 @@ time(Goal) :-
         fail
     ).
 
-report(t(OldWall, OldTime, OldInferences), Sub) :-
-    time_state(t(NewWall, NewTime, NewInferences)),
-    UsedTime is NewTime - OldTime,
-    UsedInf  is NewInferences - OldInferences - Sub,
-    Wall     is NewWall - OldWall,
-    (   UsedTime =:= 0
-    ->  Lips = 'Infinite'
-    ;   Lips is integer(UsedInf / UsedTime)
-    ),
-    print_message(information, time(UsedInf, UsedTime, Wall, Lips)).
+%!  call_time(:Goal, -Time:dict).
+%!  call_time(:Goal, -Time:dict, -Result).
+%
+%   Call Goal as  call/1,  unifying  Time   with  a  dict  that provides
+%   information on the resource usage. If   Goal  succeeds with a choice
+%   point, backtracking reports the time used to find the _next answer_,
+%   failure or exception. If Goal   succeeds deterministically no choice
+%   point is left open. Currently Time   contains the keys below. Future
+%   versions may provide additional keys.
+%
+%     - wall:Seconds
+%     - cpu:Seconds
+%     - inferences:Count
+%
+%   call_time/2 is defined as below. Note  that for call_time/2 the time
+%   is only available if Goal succeeds.
+%
+%       call_time(Goal, Time) :-
+%	    call_time(Goal, Time, Result),
+%	    call(Result).
+%
+%   @arg Result is one of  `true`,   `false`  or  throw(E), depending on
+%   whether or not the goal succeeded or  raised an exception. Note that
+%   Result may be called  using  call/1   to  propagate  the  failure or
+%   exception.
 
-time_state(t(Wall, Time, Inferences)) :-
+call_time(Goal, Time) :-
+    call_time(Goal, Time, Result),
+    call(Result).
+
+call_time(Goal, Time, Result) :-
+    time_state(State0),
+    (   call_cleanup(catch(Goal, E, true),
+                     Det = true),
+        time_true_used(State0, Time),
+        (   var(E)
+        ->  Result = true,
+            (   Det == true
+            ->  !
+            ;   true
+            )
+        ;   !,
+            Result = throw(E)
+        )
+    ;   time_used(State0, 11, Time),
+        Result = false
+    ).
+
+report(State0, Sub) :-
+    time_used(State0, Sub, time{wall:Wall, cpu:Time, inferences:Inferences}),
+    (   Time =:= 0
+    ->  Lips = 'Infinite'
+    ;   Lips is integer(Inferences/Time)
+    ),
+    print_message(information, time(Inferences, Time, Wall, Lips)).
+
+time_used(time{wall:OldWall, cpu:OldTime, inferences:OldInferences}, Sub,
+          time{wall:Wall, cpu:Time, inferences:Inferences}) :-
+    time_state(time{wall:NewWall, cpu:NewTime, inferences:NewInferences}),
+    Time       is NewTime - OldTime,
+    Inferences is NewInferences - OldInferences - Sub,
+    Wall       is NewWall - OldWall.
+
+time_state(time{wall:Wall, cpu:Time, inferences:Inferences}) :-
     get_time(Wall),
     statistics(cputime, Time),
     statistics(inferences, Inferences).
 
-time_true(State0) :-
-    report(State0, 12).             % leave choice-point
+time_true_report(State) :-             % leave choice-point
+    report(State, 12).
+time_true_report(State) :-
+    time_true(State).
+
+time_true_used(State, Time) :-         % leave choice-point
+    time_used(State, 12, Time).
+time_true_used(State, _) :-
+    time_true(State).
+
+
 time_true(State) :-
     get_time(Wall),
     statistics(cputime, Time),
     statistics(inferences, Inferences0),
-    plus(Inferences0, -3, Inferences),
-    nb_setarg(1, State, Wall),
-    nb_setarg(2, State, Time),
-    nb_setarg(3, State, Inferences),
+    Inferences is Inferences0 - 5,
+    nb_set_dict(wall, State, Wall),
+    nb_set_dict(cpu, State, Time),
+    nb_set_dict(inferences, State, Inferences),
     fail.
 
 
-                 /*******************************
-                 *     EXECUTION PROFILING      *
-                 *******************************/
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-This module provides a simple backward compatibility frontend on the new
-(in version 5.1.10) execution profiler  with  a   hook  to  the  new GUI
-visualiser for profiling results defined in library('swi/pce_profile').
-
-Later we will add a proper textual report-generator.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-:- multifile
-    prolog:show_profile_hook/1.
-
-%!  profile(:Goal).
-%!  profile(:Goal, +Options).
-%
-%   Run Goal under the execution profiler.  Defined options are:
-%
-%     * time(Which)
-%     Profile =cpu= or =wall= time.  The default is CPU time.
-%     * top(N)
-%     When generating a textual report, show the top N predicates.
-%     * cumulative(Bool)
-%     If =true= (default =false=), show cumulative output in
-%     a textual report.
-
-profile(Goal) :-
-    profile(Goal, []).
-
-profile(Goal0, Options) :-
-    option(time(Which), Options, cpu),
-    time_name(Which, How),
-    expand_goal(Goal0, Goal),
-    call_cleanup('$profile'(Goal, How),
-                 prolog_statistics:show_profile(Options)).
-
-time_name(cpu,      cputime)  :- !.
-time_name(wall,     walltime) :- !.
-time_name(cputime,  cputime)  :- !.
-time_name(walltime, walltime) :- !.
-time_name(Time, _) :-
-    must_be(oneof([cpu,wall]), Time).
-
-%!  show_profile(+Options)
-%
-%   Display last collected profiling data.  Options are
-%
-%     * top(N)
-%     When generating a textual report, show the top N predicates.
-%     * cumulative(Bool)
-%     If =true= (default =false=), show cumulative output in
-%     a textual report.
-
-show_profile(N) :-
-    integer(N),
-    !,
-    show_profile([top(N)]).
-show_profile(Options) :-
-    profiler(Old, false),
-    show_profile_(Options),
-    profiler(_, Old).
-
-show_profile_(Options) :-
-    prolog:show_profile_hook(Options),
-    !.
-show_profile_(Options) :-
-    prof_statistics(Stat),
-    sort_on(Options, SortKey),
-    findall(Node, profile_procedure_data(_:_, Node), Nodes),
-    sort_prof_nodes(SortKey, Nodes, Sorted),
-    format('~`=t~69|~n'),
-    format('Total time: ~3f seconds~n', [Stat.time]),
-    format('~`=t~69|~n'),
-    format('~w~t~w =~45|~t~w~60|~t~w~69|~n',
-           [ 'Predicate', 'Box Entries', 'Calls+Redos', 'Time'
-           ]),
-    format('~`=t~69|~n'),
-    option(top(N), Options, 25),
-    show_plain(Sorted, N, Stat, SortKey).
-
-sort_on(Options, ticks_self) :-
-    option(cumulative(false), Options, false),
-    !.
-sort_on(_, ticks).
-
-sort_prof_nodes(ticks, Nodes, Sorted) :-
-    !,
-    map_list_to_pairs(key_ticks, Nodes, Keyed),
-    sort(1, >=, Keyed, KeySorted),
-    pairs_values(KeySorted, Sorted).
-sort_prof_nodes(Key, Nodes, Sorted) :-
-    sort(Key, >=, Nodes, Sorted).
-
-key_ticks(Node, Ticks) :-
-    Ticks is Node.ticks_self + Node.ticks_siblings.
-
-show_plain([], _, _, _).
-show_plain(_, 0, _, _) :- !.
-show_plain([H|T], N, Stat, Key) :-
-    show_plain(H, Stat, Key),
-    N2 is N - 1,
-    show_plain(T, N2, Stat, Key).
-
-show_plain(Node, Stat, Key) :-
-    value(label,                       Node, Pred),
-    value(call,                        Node, Call),
-    value(redo,                        Node, Redo),
-    value(time(Key, percentage, Stat), Node, Percent),
-    IntPercent is round(Percent*10),
-    Entry is Call + Redo,
-    format('~w~t~D =~45|~t~D+~55|~D ~t~1d%~69|~n',
-           [Pred, Entry, Call, Redo, IntPercent]).
-
-
-                 /*******************************
-                 *         DATA GATHERING       *
-                 *******************************/
-
-%!  profile_data(-Data) is det.
-%
-%   Gather all relevant data from profiler. This predicate may be called
-%   while profiling is active  in  which   case  it  is  suspended while
-%   collecting the data. Data is a dict providing the following fields:
-%
-%     - summary:Dict
-%       Overall statistics providing
-%       - samples:Count:
-%         Times the statistical profiler was called
-%       - ticks:Count
-%         Virtual ticks during profiling
-%       - accounting:Count
-%         Tick spent on accounting
-%       - time:Seconds
-%         Total time sampled
-%       - nodes:Count
-%         Nodes in the call graph.
-%     - nodes
-%       List of nodes.  Each node provides:
-%       - predicate:PredicateIndicator
-%       - ticks_self:Count
-%       - ticks_siblings:Count
-%       - call:Count
-%       - redo:Count
-%       - exit:Count
-%       - callers:list_of(Relative)
-%       - callees:list_of(Relative)
-%
-%    _Relative_ is a term of the shape below that represents a caller or
-%    callee. Future versions are likely to use a dict instead.
-%
-%        node(PredicateIndicator, CycleID, Ticks, TicksSiblings,
-%             Calls, Redos, Exits)
-
-profile_data(Data) :-
-    setup_call_cleanup(
-        profiler(Old, false),
-        profile_data_(Data),
-        profiler(_, Old)).
-
-profile_data_(profile{summary:Summary, nodes:Nodes}) :-
-    prof_statistics(Summary),
-    findall(Node, profile_procedure_data(_:_, Node), Nodes).
-
-%!  prof_statistics(-Node) is det.
-%
-%   Get overall statistics
-%
-%   @param Node     term of the format prof(Ticks, Account, Time, Nodes)
-
-prof_statistics(summary{samples:Samples, ticks:Ticks,
-                        accounting:Account, time:Time, nodes:Nodes}) :-
-    '$prof_statistics'(Samples, Ticks, Account, Time, Nodes).
-
-%!  profile_procedure_data(?Pred, -Data:dict) is nondet.
-%
-%   Collect data for Pred. If Pred is   unbound  data for each predicate
-%   that has profile data available is   returned.  Data is described in
-%   profile_data/1 as an element of the `nodes` key.
-
-profile_procedure_data(Pred, Node) :-
-    Node = node{predicate:Pred,
-                ticks_self:TicksSelf, ticks_siblings:TicksSiblings,
-                call:Call, redo:Redo, exit:Exit,
-                callers:Parents, callees:Siblings},
-    (   specified(Pred)
-    ->  true
-    ;   profiled_predicates(Preds),
-        member(Pred, Preds)
-    ),
-    '$prof_procedure_data'(Pred,
-                           TicksSelf, TicksSiblings,
-                           Call, Redo, Exit,
-                           Parents, Siblings).
-
-specified(Module:Head) :-
-    atom(Module),
-    callable(Head).
-
-profiled_predicates(Preds) :-
-    setof(Pred, prof_impl(Pred), Preds).
-
-prof_impl(Pred) :-
-    prof_node_id(Node),
-    node_id_pred(Node, Pred).
-
-prof_node_id(N) :-
-    prof_node_id_below(N, -).
-
-prof_node_id_below(N, Root) :-
-    '$prof_sibling_of'(N0, Root),
-    (   N = N0
-    ;   prof_node_id_below(N, N0)
-    ).
-
-node_id_pred(Node, Pred) :-
-    '$prof_node'(Node, Pred, _Calls, _Redos, _Exits, _Recur,
-                 _Ticks, _SiblingTicks).
-
-%!  value(+Key, +NodeData, -Value)
-%
-%   Obtain possible computed attributes from NodeData.
-
-value(name, Data, Name) :-
-    !,
-    predicate_sort_key(Data.predicate, Name).
-value(label, Data, Label) :-
-    !,
-    predicate_label(Data.predicate, Label).
-value(ticks, Data, Ticks) :-
-    !,
-    Ticks is Data.ticks_self + Data.ticks_siblings.
-value(time(Key, percentage, Stat), Data, Percent) :-
-    !,
-    value(Key, Data, Ticks),
-    Total = Stat.ticks,
-    Account = Stat.accounting,
-    (   Total-Account > 0
-    ->  Percent is 100 * (Ticks/(Total-Account))
-    ;   Percent is 0.0
-    ).
-value(Name, Data, Value) :-
-    Value = Data.Name.
-
-
                  /*******************************
                  *            MESSAGES          *
                  *******************************/
